@@ -18,13 +18,17 @@
   (let ((args (copy-list args)))
     (remf args :weakness)
     (remf args :synchronized)
-    (remf args :finalizer)
     (let ((hash-fun (getf args :hash-function)))
       (when hash-fun
         (assert (eq (getf args :test) 'eq))
         (remf args :hash-function)))
     (apply 'make-hash-table args)))
 
+(defun %hash-table-alist (hash-table &aux result)
+  (maphash (lambda (key value) (push (cons key value) result))
+           hash-table)
+  result)
+
 ;;; In correct code, TRULY-THE has only a performance impact and can
 ;;; be safely degraded to ordinary THE.
 (defmacro truly-the (type expr)
@@ -52,7 +56,7 @@
 
 ;;; Necessary only to placate the host compiler in %COMPILER-DEFGLOBAL.
 (defun set-symbol-global-value (sym val)
-  (error "Can't set symbol-global-value: ~S ~S" sym val))
+  (setf (symbol-value sym) val))
 
 (defun %defun (name lambda &optional inline-expansion)
   (declare (ignore inline-expansion))
@@ -86,25 +90,12 @@
 ;;; In the target SBCL, the INSTANCE type refers to a base
 ;;; implementation for compound types with lowtag
 ;;; INSTANCE-POINTER-LOWTAG. There's no way to express exactly that
-;;; concept portably, but we know that anything derived from STRUCTURE!OBJECT
-;;; is equivalent to the target INSTANCE type. Also, because we use host packages
-;;; as proxies for target packages, those too must satisfy our INSTANCEP
-;;; - even if not a subtype of (OR STANDARD-OBJECT STRUCTURE-OBJECT).
-;;; Nothing else satisfies this definition of INSTANCEP.
-;;; As a guarantee that our set of host object types is exhaustive, we add one
-;;; more constraint when self-hosted: host instances of unknown type cause failure.
-;;; Some objects manipulated by the cross-compiler like the INTERVAL struct
-;;; - which is not a STRUCTURE!OBJECT - should never be seen as literals in code.
-;;; We assert that by way of the guard function.
-#+host-quirks-sbcl
-(defun unsatisfiable-instancep (x)
-  (when (and (host-sb-kernel:%instancep x)
-             (not (target-num-p x)))
-    (bug "%INSTANCEP test on ~S" x)))
+;;; concept portably, but we can at least declare here what kind of
+;;; host objects get dumped into something that when loaded would be a
+;;; target INSTANCE.
 (deftype instance ()
-  '(or structure!object package
-    #+host-quirks-sbcl (and host-sb-kernel:instance ; optimizes out a call when false
-                            (satisfies unsatisfiable-instancep))))
+  '(or (and structure-object (not target-num))
+       package))
 (defun %instancep (x)
   (typep x 'instance))
 
@@ -117,8 +108,6 @@
   (if (symbolp x) nil (error "Called SIMPLE-FUN-P on ~S" x)))
 (defun closurep (x)
   (if (symbolp x) nil (error "Called CLOSUREP on ~S" x)))
-(defun unbound-marker-p (x)
-  (if (symbolp x) nil (error "Called UNBOUND-MARKER-P on ~S" x)))
 (defun vector-with-fill-pointer-p (x)
   (if (symbolp x) nil (error "Called VECTOR-WITH-FILL-POINTER-P on ~S" x)))
 
@@ -138,14 +127,40 @@
        (or (not (typep x 'simple-array))
            (/= (array-rank x) 1))))
 
-(defvar sb-xc:*gensym-counter* 0)
-
+;;; While cross-compiling, we do not use *GENSYM-COUNTER* to make part
+;;; of the new symbol's name.  This is for two reasons: firstly,
+;;; genesis can coalesce uninterned symbols of similar names, which
+;;; leads to significant space savings in the resulting core;
+;;; secondly, it is very hard to maintain a consistent state of
+;;; *GENSYM-COUNTER*, even if we attempt to maintain our own: the
+;;; presence of (EVAL-WHEN (:COMPILE-TOPLEVEL ...) ...) in expansions
+;;; of e.g. SB-XC:DEFMACRO, SB-XC:DEFINE-COMPILER-MACRO and so on can
+;;; lead to macro-functions of target macros on the host, which then
+;;; will GENSYM varying number of times depending on whether they are
+;;; minimally-compiled or not: the only way to guarantee identical
+;;; symbol names at the same point in the compilation is not to use
+;;; *GENSYM-COUNTER* at all.
+;;;
+;;; Not using *GENSYM-COUNTER* would not be an issue under normal
+;;; circumstances, modulo the demands of the language spec: the only
+;;; thing that matters for the typical use of GENSYMs in code is the
+;;; identity of the symbol, not its name.  However, the coalescing of
+;;; uninterned symbols by genesis introduces a potential pitfall: if a
+;;; macro-writing-macro generates symbols at first use that the
+;;; newly-defined macro will use when *it* expands, the coalescing in
+;;; genesis would lead to that inner macro no longer having the
+;;; correct expansion.  (Or equivalents, such as macros expanded
+;;; inside definitions of functions declaimed INLINE).  In the event
+;;; of truly weird behaviour of some macro leading the intrepid
+;;; maintainer to this comment: try turning off uninterned symbol
+;;; coalescing in genesis by redefining GET-UNINTERNED-SYMBOL; if
+;;; things work after that, find the offending use of GENSYM and make
+;;; sure that everything has a distinct symbol name.  (This is not an
+;;; issue for user code, which is compiled with a normal
+;;; implementation of GENSYM).
 (defun sb-xc:gensym (&optional (thing "G"))
   (declare (type string thing))
-  (let ((n sb-xc:*gensym-counter*))
-    (prog1
-        (make-symbol (concatenate 'string thing (write-to-string n :base 10 :radix nil :pretty nil)))
-      (incf sb-xc:*gensym-counter*))))
+  (make-symbol thing))
 
 ;;; These functions are needed for constant-folding.
 (defun simple-array-nil-p (object)
@@ -192,14 +207,11 @@
   (or (find-package string)
       (error "Cross-compiler bug: no package named ~S" string)))
 
-(defmacro without-package-locks (&body body)
-  `(progn ,@body))
-
 (defmacro with-single-package-locked-error ((&optional kind thing &rest format)
                                             &body body)
   (declare (ignore kind format))
-  `(let ((.dummy. ,thing))
-     (declare (ignore .dummy.))
+  `(progn
+     ,thing
      ,@body))
 
 (defun program-assert-symbol-home-package-unlocked (context symbol control)
@@ -212,11 +224,20 @@
   name)
 
 (declaim (declaration enable-package-locks disable-package-locks))
+(declaim (declaration sb-c::tlab))
+
+;;; The XC-STRICT-CL and SB-XC packages are called COMMON-LISP in the
+;;; target image.
+(defun sb-xc:package-name (package)
+  (if (or (eq package #.(find-package "SB-XC"))
+          (eq package #.(find-package "XC-STRICT-CL")))
+      "COMMON-LISP"
+      (cl:package-name package)))
 
 ;; Nonstandard accessor for when you know you have a valid package in hand.
 ;; This avoids double lookup in *PACKAGE-NAMES* in a few places.
 ;; But portably we have to just fallback to PACKAGE-NAME.
-(defun package-%name (x) (package-name x))
+(defun package-%name (x) (sb-xc:package-name x))
 
 ;;; This definition collapses SB-XC back into COMMON-LISP.
 ;;; Use CL:SYMBOL-PACKAGE if that's not the behavior you want.
@@ -228,24 +249,34 @@
   (let ((p (cl:symbol-package symbol))
         (name (string symbol)))
     (if (and p
-             (or (eq (find-symbol name "XC-STRICT-CL") symbol)
-                 (eq (find-symbol name "SB-XC") symbol)
-                 ;; OK if the name of a symbol in the host CL package
-                 ;; is found in XC-STRICT-CL, even if the symbols
-                 ;; differ.
-                 (and (find-symbol name "XC-STRICT-CL")
-                      (eq (find-symbol name "CL") symbol))))
+             (let ((xc-strict-symbol (find-symbol name #.(find-package "XC-STRICT-CL"))))
+               (and xc-strict-symbol
+                    (or (eq xc-strict-symbol symbol)
+                        (eq symbol (find-symbol name #.(find-package "SB-XC")))
+                        (eq symbol (find-symbol name #.(find-package "COMMON-LISP")))))))
         *cl-package*
         p)))
 
+(defun possibly-base-stringize (s) (coerce (the string s) 'simple-base-string))
+(defun possibly-base-stringize-to-heap (s) (coerce (the string s) 'simple-base-string))
+
 ;;; printing structures
 
 (defun default-structure-print (structure stream depth)
   (declare (ignore depth))
   (write structure :stream stream :circle t))
 
+(defun unreachable ()
+  (bug "Unreachable reached"))
+
 (in-package "SB-KERNEL")
 
+(define-symbol-macro *gc-epoch* 0)
+(deftype weak-vector () nil)
+(defun weak-vector-p (x) (declare (ignore x)) nil)
+(defun sb-thread:make-mutex (&key name) (list :mock-mutex name))
+(deftype sb-thread:mutex () '(cons (eql :mock-mutex)))
+
 ;;; These functions are required to emulate SBCL kernel functions
 ;;; in a vanilla ANSI Common Lisp cross-compilation host.
 ;;; The emulation doesn't need to be efficient, since it's needed
@@ -258,21 +289,14 @@
 ;;   - DEBUG-SOURCE, COMPILED-DEBUG-INFO, COMPILED-DEBUG-FUN-{something}
 ;;   - HEAP-ALIEN-INFO and ALIEN-{something}-TYPE
 ;;   - COMMA
-#-metaspace (defmacro wrapper-friend (x) x)
-(defun %instance-wrapper (instance)
-  (declare (notinline classoid-wrapper))
-  (classoid-wrapper (find-classoid (type-of instance))))
+(defun %instance-layout (instance)
+  (declare (notinline classoid-layout))
+  (classoid-layout (find-classoid (type-of instance))))
 (defun %instance-length (instance)
-  (declare (notinline wrapper-length))
+  (declare (notinline layout-length))
   ;; In the target, it is theoretically possible to have %INSTANCE-LENGTH
   ;; exceeed layout length, but in the cross-compiler they're the same.
-  (wrapper-length (%instance-wrapper instance)))
-(defun %raw-instance-ref/word (instance index)
-  (declare (ignore instance index))
-  (error "No such thing as raw structure access on the host"))
-(defun layout-id (x)
-  (declare (notinline sb-kernel::wrapper-id))
-  (sb-kernel::wrapper-id x))
+  (layout-length (%instance-layout instance)))
 
 (defun %find-position (item seq from-end start end key test)
   (let ((position (position item seq :from-end from-end
@@ -288,6 +312,42 @@
 (defmacro sap-ref-word (sap offset)
   `(#+64-bit sap-ref-64 #-64-bit sap-ref-32 ,sap ,offset))
 
+;;; Needed for assembler.
+(defstruct (asm-sap-wrapper (:constructor vector-sap (vector)))
+  (vector (missing-arg) :type (simple-array (unsigned-byte 8) (*))))
+(defmacro with-pinned-objects (list &body body)
+  (declare (ignore list))
+  `(progn ,@body))
+;;; The assembler does not USE-PACKAGE sb-sys, which works to our advantage
+;;; in that we can define SAP-REF-16 and -32 as macros in the ASM package
+;;; which avoids conflict with genesis. Genesis has its own SAP emulations in SB-SYS
+;;; so all the definitions of FIXUP-CODE-OBJECT using the native accessors
+;;; can transparently operate on genesis's model of target code blobs.
+(defsetf sb-assem::sap-ref-16 sb-assem::asm-set-sap-ref-16)
+(defsetf sb-assem::sap-ref-32 sb-assem::asm-set-sap-ref-32)
+(defun sb-assem::asm-set-sap-ref-16 (sap index val)
+  (declare (type asm-sap-wrapper sap))
+  (multiple-value-bind (b0 b1)
+    #+little-endian (values (ldb (byte 8 0) val) (ldb (byte 8 8) val))
+    #+big-endian    (values (ldb (byte 8 8) val) (ldb (byte 8 0) val))
+    (let ((octets (asm-sap-wrapper-vector sap)))
+      (setf (aref octets (+ index 0)) b0
+            (aref octets (+ index 1)) b1)))
+  val)
+(defun sb-assem::asm-set-sap-ref-32 (sap index val)
+  (declare (type asm-sap-wrapper sap))
+  (multiple-value-bind (b0 b1 b2 b3)
+      #+little-endian (values (ldb (byte 8  0) val) (ldb (byte 8  8) val)
+                              (ldb (byte 8 16) val) (ldb (byte 8 24) val))
+      #+big-endian    (values (ldb (byte 8 24) val) (ldb (byte 8 16) val)
+                              (ldb (byte 8  8) val) (ldb (byte 8  0) val))
+    (let ((octets (asm-sap-wrapper-vector sap)))
+      (setf (aref octets (+ index 0)) b0
+            (aref octets (+ index 1)) b1
+            (aref octets (+ index 2)) b2
+            (aref octets (+ index 3)) b3)))
+  val)
+
 (defun logically-readonlyize (x) x)
 
 ;;; Mainly for the fasl loader
@@ -302,6 +362,17 @@
 ;;;; in lieu of #+sb-xc-host elsewere which messes up toplevel form numbers.
 (in-package "SB-C")
 
+(defun allocate-weak-vector (n) (make-array (the integer n)))
+
+#+weak-vector-readbarrier
+(progn (deftype weak-vector () nil) ; nothing is a weak-vector
+       (defun sb-int:weak-vector-ref (v i)
+         (error "Called WEAK-VECTOR-REF on ~S ~S" v i))
+       (defun (setf sb-int:weak-vector-ref) (new v i)
+         (error "Called (SETF WEAK-VECTOR-REF) on ~S ~S ~S" new v i))
+       (defun sb-int:weak-vector-len (v)
+         (error "Called WEAK-VECTOR-LEN on ~S" v)))
+
 ;;; For macro lambdas that are processed by the host
 (declaim (declaration top-level-form))
 
@@ -355,3 +426,28 @@
   ;; Let's suppose that the host lisp knows what it's doing to efficiently
   ;; compile the standard macro. Don't try to outdo it.
   `(with-output-to-string (,var) ,@body))
+
+(defun source-location ())
+
+;;; %SYMBOL-INFO is a primitive object accessor defined in 'objdef.lisp'
+;;; But in the host Lisp, there is no such thing. Instead, SYMBOL-%INFO
+;;; is kept as a property on the host symbol.
+(declaim (inline symbol-%info))
+(defun symbol-%info (symbol) (get symbol :sb-xc-globaldb-info))
+(defun symbol-dbinfo (symbol) (symbol-%info symbol))
+
+(defun sys-copy-struct (x) (copy-structure x))
+(defun ensure-heap-list (x) (copy-list x))
+
+(defun range< (l x h) (< l x h))
+(defun range<= (l x h) (<= l x h))
+(defun range<<= (l x h) (and (< l x) (<= x h)))
+(defun range<=< (l x h) (and (<= l x) (< x h)))
+
+(defvar *unbound-marker* (make-symbol "UNBOUND-MARKER"))
+
+(defun make-unbound-marker ()
+  *unbound-marker*)
+
+(defun unbound-marker-p (x)
+  (eq x *unbound-marker*))
