@@ -16,14 +16,22 @@
 (in-package "SB-C")
 
 ;;; ANSI limits on compilation
-(defconstant call-arguments-limit most-positive-fixnum
+;;; On AMD64 we prefer to use the ECX (not RCX) register,
+;;; which means that there can only be 32 bits of precision,
+;;; so accounting for the fixnum tag and 1 bit for the sign,
+;;; this leaves 30 bits. Of course this number is ridiculous
+;;; as a call with that many args would consume 8 GB of stack,
+;;; but it's surely not as ridiculous as MOST-POSITIVE-FIXNUM.
+(defconstant call-arguments-limit
+  #+x86-64 (ash 1 30)
+  #-x86-64 most-positive-fixnum
   "The exclusive upper bound on the number of arguments which may be passed
   to a function, including &REST args.")
-(defconstant lambda-parameters-limit most-positive-fixnum
+(defconstant lambda-parameters-limit call-arguments-limit
   "The exclusive upper bound on the number of parameters which may be specified
   in a given lambda list. This is actually the limit on required and &OPTIONAL
   parameters. With &KEY and &AUX you can get more.")
-(defconstant multiple-values-limit most-positive-fixnum
+(defconstant multiple-values-limit call-arguments-limit
   "The exclusive upper bound on the number of multiple VALUES that you can
   return.")
 
@@ -96,60 +104,67 @@
 
 ;;;; miscellaneous utilities
 
-(defstruct (debug-name-marker (:print-function print-debug-name-marker)
-                              (:copier nil)))
-(declaim (freeze-type debug-name-marker))
+;;; COMPILE-FILE usually puts all nontoplevel code in immobile space, but COMPILE
+;;; offers a choice. Because the immobile space GC does not run often enough (yet),
+;;; COMPILE usually places code in the dynamic space managed by our copying GC.
+;;; Change this variable if your application always demands immobile code.
+;;; In particular, ELF cores shrink the immobile code space down to just enough
+;;; to contain all code, plus about 1/2 MiB of spare, which means that you can't
+;;; subsequently compile a whole lot into immobile space.
+;;; The value is changed to :AUTO in make-target-2-load.lisp which supresses
+;;; codegen optimizations for immobile space, but nonetheless prefers to allocate
+;;; the code there, falling back to dynamic space if there is no room left.
+;;; These controls exist whether or not the immobile-space feature is present.
+(declaim (type (member :immobile :dynamic :auto) *compile-to-memory-space*)
+         (type (member :immobile :dynamic) *compile-file-to-memory-space*))
+(defvar *compile-to-memory-space* :immobile) ; BUILD-TIME default
+(export '*compile-file-to-memory-space*) ; silly user code looks at, even if no immobile-space
+(defvar *compile-file-to-memory-space* :immobile) ; BUILD-TIME default
 
-(defvar *debug-name-level* 4)
-(defvar *debug-name-length* 12)
-(defvar *debug-name-punt*)
-(define-load-time-global *debug-name-sharp* (make-debug-name-marker))
-(define-load-time-global *debug-name-ellipsis* (make-debug-name-marker))
-
-(defun print-debug-name-marker (marker stream level)
-  (declare (ignore level))
-  (cond ((eq marker *debug-name-sharp*)
-         (write-char #\# stream))
-        ((eq marker *debug-name-ellipsis*)
-         (write-string "..." stream))
-        (t
-         (write-string "???" stream))))
+(defun compile-perfect-hash (lambda test-inputs)
+  ;; Don't blindly trust the hash generator: assert that computed values are
+  ;; in range and not repeated.
+  (let ((seen (make-array (power-of-two-ceiling (length test-inputs))
+                          :element-type 'bit :initial-element 0))
+        (f #-sb-xc-host ; use fasteval if possible
+           (cond #+sb-fasteval
+                 ((< (length test-inputs) 100) ; interpreting is faster
+                  (let ((*evaluator-mode* :interpret))
+                    (eval lambda)))
+                 (t (let ((*compile-to-memory-space* :dynamic))
+                      (compile nil lambda))))
+           #+sb-xc-host
+           (destructuring-bind (head lambda-list . body) lambda
+             (aver (eq head 'lambda))
+             (multiple-value-bind (forms decls) (parse-body body nil)
+               (declare (ignore decls))
+               ;; Give the host a definition for SB-C::UINT32-MODULARLY and remove _all_
+               ;; OPTIMIZE decls hidden within. We don't need to pedantically correctly
+               ;; code-walk here, because hash expressions are largely boilerplate that
+               ;; will not confusingly match forms such as (LET ((OPTIMIZE ...))).
+               (let ((new-body
+                      `(macrolet ((uint32-modularly (&whole form &rest exprs)
+                                    (declare (ignore exprs))
+                                    (funcall (sb-xc:macro-function 'sb-c::uint32-modularly)
+                                             form nil)))
+                         ,@(subst-if '(optimize)
+                                     (lambda (x) (typep x '(cons (eql optimize) (not null))))
+                                     forms))))
+                 (compile nil `(lambda ,lambda-list ,new-body)))))))
+    (loop for input across test-inputs
+          do (let ((h (funcall f input)))
+               (unless (zerop (bit seen h))
+                 (bug "Perfect hash generator failed on ~X" test-inputs))
+               (setf (bit seen h) 1)))
+    f))
 
 (declaim (ftype (sfunction () list) name-context))
 (defun debug-name (type thing &optional context)
-  (let ((*debug-name-punt* nil))
-    (labels ((walk (x)
-               (typecase x
-                 (cons
-                  (if (plusp *debug-name-level*)
-                      (let ((*debug-name-level* (1- *debug-name-level*)))
-                        (do ((tail (cdr x) (cdr tail))
-                             (name (cons (walk (car x)) nil)
-                                   (cons (walk (car tail)) name))
-                             (n (1- *debug-name-length*) (1- n)))
-                            ((or (not (consp tail))
-                                 (not (plusp n))
-                                 *debug-name-punt*)
-                             (cond (*debug-name-punt*
-                                    (setf *debug-name-punt* nil)
-                                    (nreverse name))
-                                   ((atom tail)
-                                    (nconc (nreverse name) (walk tail)))
-                                   (t
-                                    (setf *debug-name-punt* t)
-                                    (nconc (nreverse name) (list *debug-name-ellipsis*)))))))
-                      *debug-name-sharp*))
-                 ((or symbol number string)
-                  x)
-                 (t
-                  ;; wtf?? This looks like a source of sensitivity to the cross-compiler host
-                  ;; in addition to which it seems generally a stupid idea.
-                  (type-of x)))))
-      (let ((name (list* type (walk thing) (when context (name-context)))))
-        (when (legal-fun-name-p name)
-          (bug "~S is a legal function name, and cannot be used as a ~
-                debug name." name))
-        name))))
+  (let ((name (list* type thing (when context (name-context)))))
+    (when (legal-fun-name-p name)
+      (bug "~S is a legal function name, and cannot be used as a ~
+            debug name." name))
+    name))
 
 ;;; Bound during eval-when :compile-time evaluation.
 (defvar *compile-time-eval* nil)
@@ -191,6 +206,13 @@
     (list (make-hash-table :test 'equal :synchronized t)))
   (declaim (type (cons hash-table) *code-coverage-info*)))
 
+;;; Unique number assigned into high 4 bytes of 64-bit code size slot
+;;; so that we can sort the contents of text space in a more-or-less
+;;; predictable manner based on the order in which code was loaded.
+;;; This wraps around at 32 bits, but it's still deterministic.
+(define-load-time-global *code-serialno* 0)
+(declaim (fixnum *code-serialno*))
+
 (deftype id-array ()
   '(and (array t (*))
         ;; Might as well be as specific as we can.
@@ -213,6 +235,13 @@
   (coverage-metadata nil :type (or (cons hash-table hash-table) null) :read-only t)
   (msan-unpoison nil :read-only t)
   (sset-counter 1 :type fixnum)
+  ;; Map of function name -> something about how many calls were converted
+  ;; as ordinary calls not in the scope of a local or global notinline declaration.
+  ;; Useful for finding functions that were supposed to have been converted
+  ;; through some kind of transformation but were not.
+  ;; FIXME: this should be scoped to a compile/load but there are
+  ;; apparently some difficulties in doing so.
+  ; (emitted-full-calls (make-hash-table :test 'equal))
   ;; if emitting a cfasl, the fasl stream to that
   (compile-toplevel-object nil :read-only t)
   ;; The current block compilation state.  These are initialized to
@@ -257,3 +286,39 @@
 ;;  "#define MEM_TO_SHADOW(mem) (((uptr)(mem)) ^ 0x500000000000ULL)"
 #+linux ; shadow space differs by OS
 (defconstant sb-vm::msan-mem-to-shadow-xor-const #x500000000000)
+
+(define-load-time-global *emitted-full-calls*
+    (make-hash-table :test 'equal #-sb-xc-host :synchronized #-sb-xc-host t))
+
+(defmacro get-emitted-full-calls (name)
+;; Todo: probably remove the wrapping cons. It was for globaldb
+;; which is particularly inefficient at updates (because it can only
+;; use an R/C/U paradigm, and so conses on every insert,
+;; unlike a hash-table which can just update the cell)
+  `(gethash ,name *emitted-full-calls*))
+
+;; Return the number of calls to NAME that IR2 emitted as full calls,
+;; not counting calls via #'F that went untracked.
+;; Return 0 if the answer is nonzero but a warning was already signaled
+;; about any full calls were emitted. This return convention satisfies the
+;; intended use of this statistic - to decide whether to generate a warning
+;; about failure to inline NAME, which is shown at most once per name
+;; to avoid unleashing a flood of identical warnings.
+(defun emitted-full-call-count (name)
+  (let ((status (get-emitted-full-calls name)))
+    (and (integerp status)
+         ;; Bit 0 tells whether any call was NOT in the presence of
+         ;; a 'notinline' declaration, thus eligible to be inline.
+         ;; Bit 1 tells whether any warning was emitted yet.
+         (= (logand status 3) #b01)
+         (ash status -2)))) ; the call count as tracked by IR2
+
+(defun accumulate-full-calls (data)
+  (loop for (name status) in data
+        do
+        (let ((existing (gethash name *emitted-full-calls* 0)))
+          (setf (gethash name *emitted-full-calls*)
+                (logior (+ (logand existing #b11) ; old flag bits
+                           (logand status #b11))  ; new flag bits
+                        (logand existing -4)      ; old count
+                        (logand status -4))))))   ; new count
