@@ -85,6 +85,11 @@
 (defconstant array-rank-limit 129
   "the exclusive upper bound on the rank of an array")
 
+(defconstant max-dynamic-space-size
+  (or #+soft-card-marks
+      (* (/ (expt 2 31) sb-vm::cards-per-page) sb-c:+backend-page-bytes+)
+      (1- (expt 2 sb-vm:n-word-bits))))
+
 ;;; FIXME: these limits are wrong at the most basic level according to the spec,
 ;;; because if there are different limits for different element types, then
 ;;; this constant should be the _smallest_ of the limits for any element type.
@@ -95,10 +100,13 @@
 #-ubsan ; usual way
 (progn
 ;;; - 2 to leave space for the array header
-(defconstant array-dimension-limit (- most-positive-fixnum 2)
+(defconstant array-dimension-limit (min (* max-dynamic-space-size
+                                           ;; for bit-arrays
+                                           n-byte-bits)
+                                        (- most-positive-fixnum 2))
   "the exclusive upper bound on any given dimension of an array")
 
-(defconstant array-total-size-limit (- most-positive-fixnum 2)
+(defconstant array-total-size-limit array-dimension-limit
   "the exclusive upper bound on the total number of elements in an array")
 )
 
@@ -124,18 +132,18 @@
 ;;; running full tilt will exhaust this faster, but it's still plenty enough.
 (defconstant sb-kernel::internal-time-bits 61)
 
-(defconstant most-positive-exactly-single-float-fixnum
-  (min (expt 2 single-float-digits) most-positive-fixnum))
-(defconstant most-negative-exactly-single-float-fixnum
-  (max (- (expt 2 single-float-digits)) most-negative-fixnum))
-(defconstant most-positive-exactly-double-float-fixnum
-  (min (expt 2 double-float-digits) most-positive-fixnum))
-(defconstant most-negative-exactly-double-float-fixnum
-  (max (- (expt 2 double-float-digits)) most-negative-fixnum))
+(defconstant most-positive-exactly-single-float-integer
+  (expt 2 single-float-digits))
+(defconstant most-negative-exactly-single-float-integer
+  (- (expt 2 single-float-digits)))
+(defconstant most-positive-exactly-double-float-integer
+  (expt 2 double-float-digits))
+(defconstant most-negative-exactly-double-float-integer
+  (- (expt 2 double-float-digits)))
 
 ;;;; Point where continuous area starting at dynamic-space-start bumps into
 ;;;; next space. Computed for genesis/constants.h, not used in Lisp.
-#+(and gencgc sb-xc-host)
+#+(and generational sb-xc-host)
 (defconstant max-dynamic-space-end
     (let ((stop (1- (ash 1 n-word-bits)))
           (start dynamic-space-start))
@@ -162,7 +170,7 @@
 ;;; which can be expressed in 8 bits.
 (defconstant short-header-max-words #x7fff)
 
-#+gencgc
+#+generational
 (progn
   (defconstant gencgc-page-words (/ gencgc-page-bytes n-word-bytes))
   ;; Preventing use of the last 2 words on a page ensures that we never
@@ -188,4 +196,9 @@
 ;;; END of the global safepoint page.
 (defconstant gc-safepoint-trap-offset n-word-bytes)
 
-#+sb-xc-host (deftype sb-xc:fixnum () `(signed-byte ,n-fixnum-bits))
+#+sb-xc-host
+(progn
+  (deftype sb-xc:fixnum () `(signed-byte ,n-fixnum-bits))
+  ;; %char-code seems to belong in 'cross-char' but our CHAR-CODE-LIMIT
+  ;; is not defined by then.
+  (deftype %char-code () `(integer 0 (,sb-xc:char-code-limit))))
