@@ -40,8 +40,7 @@
 ;; By happenstance this is the same as small-space-size.
 (defconstant alien-linkage-table-space-size #x100000)
 
-#+gencgc
-;; Define START/END constants for GENCGC spaces.
+;; Define START/END constants for GC spaces.
 ;; Assumptions:
 ;;     We only need very small read-only and static spaces, because
 ;;     gencgc does not purify any more.  We can count on being able to
@@ -55,7 +54,7 @@
 ;;
 ;;     The safepoint page (if enabled) is to be allocated immediately
 ;;     prior to static page.
-(defmacro !gencgc-space-setup
+(defmacro gc-space-setup
     (small-spaces-start
           ;; These keywords variables have to be careful not to overlap with the
           ;; the DEFCONSTANT of the same name, hence the suffix.
@@ -65,9 +64,9 @@
                ;; except in forcing discontiguous addresses for testing.
                ;; And of course, don't use them if unsupported.
                ((:fixedobj-space-start fixedobj-space-start*))
-               ((:fixedobj-space-size  fixedobj-space-size*) (* 24 1024 1024))
+               ((:fixedobj-space-size  fixedobj-space-size*) (* 48 1024 1024))
                ((:text-space-start text-space-start*))
-               ((:text-space-size  text-space-size*) (* 104 1024 1024))
+               ((:text-space-size  text-space-size*) (* 160 1024 1024))
                (small-space-size #x100000)
                ((:read-only-space-size ro-space-size)
                 #+darwin-jit small-space-size
@@ -87,10 +86,13 @@
                            #+(and win32 x86-64)
                            (seh-data ,(symbol-value '+backend-page-bytes+) win64-seh-data-addr)
                            #-immobile-space (alien-linkage-table ,alien-linkage-table-space-size)
-                           #+sb-safepoint
+                           ;; safepoint on 64-bit uses a relocatable trap page just below the card mark
+                           ;; table, which works nicely assuming a register is wired to the card table
+                           #+(and sb-safepoint (not x86-64))
                            ;; Must be just before NIL.
-                           (safepoint ,(symbol-value '+backend-page-bytes+) gc-safepoint-page-addr)
+                           (safepoint ,(symbol-value '+backend-page-bytes+))
                            (static ,small-space-size)
+                           (permgen 8388608) ; 8MiB
                            #+darwin-jit
                            (static-code ,small-space-size))
                          #+immobile-space
@@ -102,10 +104,11 @@
            (loop for (space size var-name) in spaces
                  appending
                  (let* ((relocatable
-                          ;; READONLY is usually movable now.
-                          (member space '(fixedobj text
+                          (member space '(fixedobj text permgen
+                                          #+relocatable-static-space safepoint
+                                          #+relocatable-static-space static
                                           #+immobile-space alien-linkage-table
-                                          #-darwin-jit read-only)))
+                                          read-only)))
                         (start ptr)
                         (end (+ ptr size)))
                    (setf ptr end)
@@ -140,7 +143,7 @@
                  (32 (expt 2 29))
                  (64 (expt 2 30)))))
          ;; an arbitrary value to avoid kludging genesis
-         #+(and sb-xc-host (not darwin-jit))
+         #+sb-xc-host
          (defparameter read-only-space-end read-only-space-start)
          #-soft-card-marks (defconstant cards-per-page 1)
          (defconstant gencgc-card-bytes (/ gencgc-page-bytes cards-per-page))
@@ -158,6 +161,8 @@
          (defconstant gencgc-card-table-index-mask
            (1- (ash 1 gencgc-card-table-index-nbits)))))))
 
+(defconstant card-marked #+soft-card-marks 0 #-soft-card-marks 1)
+
 (defconstant-eqx +c-callable-fdefns+
   '(sub-gc
     sb-kernel::post-gc
@@ -187,22 +192,24 @@
 ;;; On #+sb-thread builds, these are not static, because access to them
 ;;; is via the TLS, not the symbol.
 (defconstant-eqx per-thread-c-interface-symbols
-  `((*free-interrupt-context-index* 0)
-    (sb-sys:*allow-with-interrupts* t)
-    (sb-sys:*interrupts-enabled* t)
-    sb-sys:*interrupt-pending*
-    #+sb-safepoint sb-sys:*thruption-pending*
-    *in-without-gcing*
-    *gc-inhibit*
-    *gc-pending*
-    #+sb-safepoint sb-impl::*in-safepoint*
-    #+sb-thread *stop-for-gc-pending*
-    *pinned-objects*
-    #+gencgc (*gc-pin-code-pages* 0)
-    ;; things needed for non-local-exit
-    (*current-catch-block* 0)
-    (*current-unwind-protect-block* 0)
-    )
+    (hash-cons
+     '((*free-interrupt-context-index* 0)
+       (sb-sys:*allow-with-interrupts* t)
+       (sb-sys:*interrupts-enabled* t)
+       sb-sys:*interrupt-pending*
+       #+sb-safepoint sb-sys:*thruption-pending*
+       *in-without-gcing*
+       *gc-inhibit*
+       *gc-pending*
+       #+sb-safepoint sb-impl::*in-safepoint*
+       #+sb-thread *stop-for-gc-pending*
+       sb-impl::*unweakened-vectors*
+       *pinned-objects*
+       (*gc-pin-code-pages* 0)
+       ;; things needed for non-local-exit
+       #+ultrafutex (*current-mutex* 0)
+       (*current-catch-block* 0)
+       (*current-unwind-protect-block* 0)))
   #'equal)
 
 (defconstant-eqx +common-static-symbols+
@@ -218,13 +225,13 @@
     ,@'(*current-catch-block*
         *current-unwind-protect-block*)
 
-    #+metaspace *metaspace-tracts*
     *immobile-codeblob-tree* ; for generations 0 through 5 inclusive
     *immobile-codeblob-vector* ; for pseudo-static-generation
     *dynspace-codeblob-tree*
     sb-impl::**finalizer-store**
     sb-impl::*finalizer-rehashlist*
     sb-impl::*finalizers-triggered*
+    sb-impl::*run-gc-hooks*
 
     ;; stack pointers
     #-sb-thread *binding-stack-start* ; a thread slot if #+sb-thread
@@ -237,6 +244,9 @@
     #+sb-thread sb-thread::*starting-threads*
     *free-tls-index* ; always exists for benefit of C runtime
 
+    #+(and x86-64 sb-thread (not gs-seg))
+    sb-aprof::*n-profile-sites*
+
     ;; runtime linking of lisp->C calls (regardless of whether
     ;; the C function is in a dynamic shared object or not)
     +required-foreign-symbols+
@@ -320,6 +330,12 @@
       ;; fdefn pointers in the core. A couple others were high on the list as well.
       hairy-data-vector-set
       hairy-data-vector-ref
+      vector-hairy-data-vector-set
+      vector-hairy-data-vector-ref
+      hairy-data-vector-set/check-bounds
+      hairy-data-vector-ref/check-bounds
+      vector-hairy-data-vector-set/check-bounds
+      vector-hairy-data-vector-ref/check-bounds
       %ldb
       sb-kernel:vector-unsigned-byte-8-p)
   #'equalp)
@@ -346,20 +362,21 @@
   #-sb-thread 1024) ; crazy value
 
 ;;; Thread slots accessed at negative indices relative to struct thread.
+;;; FIXME: this is extremely unmaintainable.
 (defconstant thread-header-slots
   ;; This seems to need to be an even number.
   ;; I'm not sure what the constraint on that stems from.
   #+(and x86-64 sb-safepoint) 14 ; the safepoint trap page is at word index -15
   #+(and x86-64 (not sb-safepoint)) 16
-  #-x86-64 0)
+  #+(and (not x86-64) immobile-space) 14 ; the safepoint trap page is at word index -15
+  #+(and (not x86-64) (not immobile-space)) 0)
 
-#+gencgc
 (progn
   (defconstant +highest-normal-generation+ 5)
   (defconstant +pseudo-static-generation+ 6))
 
 (defparameter *runtime-asm-routines* nil)
-(defparameter *linkage-space-predefined-entries* nil)
+(defparameter *alien-linkage-table-predefined-entries* nil)
 
 ;;; Floating-point related constants, both format descriptions and FPU
 ;;; control register descriptions.  These don't exactly match up with
@@ -390,18 +407,28 @@
 (defconstant single-float-hidden-bit (ash 1 23))
 
 (defconstant double-float-bias 1022)
-(defconstant-eqx double-float-exponent-byte (byte 11 20) #'equalp)
-(defconstant-eqx double-float-significand-byte (byte 20 0) #'equalp)
+(defconstant-eqx double-float-exponent-byte (byte 11 52) #'equalp)
+(defconstant-eqx double-float-significand-byte (byte 52 0) #'equalp)
+(defconstant-eqx double-float-hi-exponent-byte (byte 11 20) #'equalp)
+(defconstant-eqx double-float-hi-significand-byte (byte 20 0) #'equalp)
 (defconstant double-float-normal-exponent-min 1)
 (defconstant double-float-normal-exponent-max #x7FE)
-(defconstant double-float-hidden-bit (ash 1 20))
+(defconstant double-float-hidden-bit (ash 1 52))
 
-(defconstant single-float-digits
-  (+ (byte-size single-float-significand-byte) 1))
-
-(defconstant double-float-digits
-  (+ (byte-size double-float-significand-byte) 32 1))
+(defconstant single-float-digits 24)
+(defconstant double-float-digits 53)
 )
 
+;;; Reserve some bits of SYMBOL-HASH slot for future use
+#+64-bit
+(progn (defconstant n-symbol-hash-prng-bits 10) ; how many to randomize
+       (defconstant n-symbol-hash-discard-bits
+         (let ((precision (+ 32 n-symbol-hash-prng-bits))) ; total N bits
+           (- 64 precision)))
+       (defconstant-eqx sb-impl::symbol-hash-prng-byte
+         (byte n-symbol-hash-prng-bits (- 32 n-symbol-hash-prng-bits))
+         #'equal))
+#-64-bit (defconstant-eqx sb-impl::symbol-hash-prng-byte (byte 3 0) #'equal)
+
 (push '("SB-VM" +c-callable-fdefns+ +common-static-symbols+)
       *!removable-symbols*)
