@@ -75,7 +75,6 @@
 
 (defun emit-inits (node block object lowtag inits args)
   (let ((unbound-marker-tn nil)
-        (funcallable-instance-tramp-tn nil)
         (dx-p (node-stack-allocate-p node)))
     (flet ((zero-init-p (x)
              ;; dynamic-space is already zeroed
@@ -136,14 +135,7 @@
                                     (vop make-unbound-marker node block tn)
                                     tn))))
                        (:null
-                        (emit-constant nil))
-                       (:funcallable-instance-tramp
-                        (or funcallable-instance-tramp-tn
-                            (setf funcallable-instance-tramp-tn
-                                  (let ((tn (make-restricted-tn
-                                             nil sb-vm:any-reg-sc-number)))
-                                    (vop make-funcallable-instance-tramp node block tn)
-                                    tn)))))
+                        (emit-constant nil)))
                      :allocator slot lowtag))))))))
   (unless (null args)
     (bug "Leftover args: ~S" args))
@@ -221,9 +213,6 @@
         (link-allocator-to-inits allocator-vop last-init-vop node))
       (move-lvar-result node block locs lvar))))
 
-;;; FIXME: this causes emission of GC store barriers, but it should not.
-;;; The vector is freshly consed, so anything being stored into it
-;;; is at least as old.
 (defoptimizer (initialize-vector ir2-convert)
     ((vector &rest initial-contents) node block)
   (let* ((vector-ctype (lvar-type vector))
@@ -369,7 +358,11 @@
                                  (if (sb-c:msan-unpoison sb-c:*compilation*)
                                      'sb-vm::allocate-vector-on-stack+msan-unpoison
                                      'sb-vm::allocate-vector-on-stack)
-                                 'sb-vm::allocate-vector-on-heap))
+                                 (cond #-x86-64
+                                       ((sb-vm::system-tlab-p 0 call)
+                                        'sb-vm::sys-allocate-vector-on-heap)
+                                       (t
+                                        'sb-vm::allocate-vector-on-heap))))
 
 (defun make-vector-check-overflow-p (node)
   (not (or (zerop (policy node safety))
@@ -411,6 +404,9 @@
 (defoptimizer (%make-complex stack-allocate-result) ((&rest args))
   t)
 
+(defoptimizer (unaligned-dx-cons stack-allocate-result) ((car))
+  t)
+
 ;;; MAKE-LIST optimizations
 #+x86-64
 (progn
