@@ -9,12 +9,7 @@
 ;;;; provided with absolutely no warranty. See the COPYING and CREDITS
 ;;;; files for more information.
 
-(in-package "SB!VM")
-
-(defun my-make-wired-tn (prim-type-name sc-name offset)
-  (make-wired-tn (primitive-type-or-lose prim-type-name)
-                 (sc-number-or-lose sc-name)
-                 offset))
+(in-package "SB-VM")
 
 (defstruct arg-state
   (stack-frame-size 0)
@@ -28,11 +23,11 @@
     (multiple-value-bind
         (ptype reg-sc stack-sc)
         (if (alien-integer-type-signed type)
-            (values 'signed-byte-32 'signed-reg 'signed-stack)
-            (values 'unsigned-byte-32 'unsigned-reg 'unsigned-stack))
+            (values 'signed-byte-32 signed-reg-sc-number signed-stack-sc-number)
+            (values 'unsigned-byte-32 unsigned-reg-sc-number unsigned-stack-sc-number))
       (if (< stack-frame-size 4)
-          (my-make-wired-tn ptype reg-sc (+ stack-frame-size 4))
-          (my-make-wired-tn ptype stack-sc stack-frame-size)))))
+          (make-wired-tn* ptype reg-sc (+ stack-frame-size 4))
+          (make-wired-tn* ptype stack-sc stack-frame-size)))))
 
 (define-alien-type-method (system-area-pointer :arg-tn) (type state)
   (declare (ignore type))
@@ -40,11 +35,9 @@
     (setf (arg-state-stack-frame-size state) (1+ stack-frame-size))
     (setf (arg-state-did-int-arg state) t)
     (if (< stack-frame-size 4)
-        (my-make-wired-tn 'system-area-pointer
-                          'sap-reg
+        (make-wired-tn* 'system-area-pointer sap-reg-sc-number
                           (+ stack-frame-size 4))
-        (my-make-wired-tn 'system-area-pointer
-                          'sap-stack
+        (make-wired-tn* 'system-area-pointer sap-stack-sc-number
                           stack-frame-size))))
 
 (define-alien-type-method (double-float :arg-tn) (type state)
@@ -54,17 +47,14 @@
     (setf (arg-state-stack-frame-size state) (+ stack-frame-size 2))
     (setf (arg-state-float-args state) (1+ float-args))
     (cond ((>= stack-frame-size 4)
-           (my-make-wired-tn 'double-float
-                             'double-stack
+           (make-wired-tn* 'double-float double-stack-sc-number
                              stack-frame-size))
           ((and (not (arg-state-did-int-arg state))
                 (< float-args 2))
-           (my-make-wired-tn 'double-float
-                             'double-reg
+           (make-wired-tn* 'double-float double-reg-sc-number
                              (+ (* float-args 2) 12)))
           (t
-           (my-make-wired-tn 'double-float
-                             'double-int-carg-reg
+           (make-wired-tn* 'double-float double-int-carg-reg-sc-number
                              (+ stack-frame-size 4))))))
 
 (define-alien-type-method (single-float :arg-tn) (type state)
@@ -74,17 +64,14 @@
     (setf (arg-state-stack-frame-size state) (1+ stack-frame-size))
     (setf (arg-state-float-args state) (1+ float-args))
     (cond ((>= stack-frame-size 4)
-           (my-make-wired-tn 'single-float
-                             'single-stack
+           (make-wired-tn* 'single-float single-stack-sc-number
                              stack-frame-size))
           ((and (not (arg-state-did-int-arg state))
                 (< float-args 2))
-           (my-make-wired-tn 'single-float
-                             'single-reg
+           (make-wired-tn* 'single-float single-reg-sc-number
                              (+ (* float-args 2) 12)))
           (t
-           (my-make-wired-tn 'single-float
-                             'single-int-carg-reg
+           (make-wired-tn* 'single-float single-int-carg-reg-sc-number
                              (+ stack-frame-size 4))))))
 
 (defstruct result-state
@@ -100,28 +87,28 @@
     (setf (result-state-num-results state) (1+ num-results))
     (multiple-value-bind (ptype reg-sc)
         (if (alien-integer-type-signed type)
-            (values 'signed-byte-32 'signed-reg)
-            (values 'unsigned-byte-32 'unsigned-reg))
-      (my-make-wired-tn ptype reg-sc (result-reg-offset num-results)))))
+            (values 'signed-byte-32 signed-reg-sc-number)
+            (values 'unsigned-byte-32 unsigned-reg-sc-number))
+      (make-wired-tn* ptype reg-sc (result-reg-offset num-results)))))
 
 (define-alien-type-method (system-area-pointer :result-tn) (type state)
   (declare (ignore type))
   (let ((num-results (result-state-num-results state)))
     (setf (result-state-num-results state) (1+ num-results))
-    (my-make-wired-tn 'system-area-pointer 'sap-reg (result-reg-offset num-results))))
+    (make-wired-tn* 'system-area-pointer sap-reg-sc-number (result-reg-offset num-results))))
 
 ;;; FIXME: do these still work? -- CSR, 2002-08-28
 (define-alien-type-method (double-float :result-tn) (type state)
   (declare (ignore type))
   (let ((num-results (result-state-num-results state)))
     (setf (result-state-num-results state) (1+ num-results))
-    (my-make-wired-tn 'double-float 'double-reg (* num-results 2))))
+    (make-wired-tn* 'double-float double-reg-sc-number (* num-results 2))))
 
 (define-alien-type-method (single-float :result-tn) (type state)
   (declare (ignore type))
   (let ((num-results (result-state-num-results state)))
     (setf (result-state-num-results state) (1+ num-results))
-    (my-make-wired-tn 'single-float 'single-reg (* num-results 2))))
+    (make-wired-tn* 'single-float single-reg-sc-number (* num-results 2))))
 
 (define-alien-type-method (values :result-tn) (type state)
   (let ((values (alien-values-type-values type)))
@@ -136,7 +123,7 @@
     (collect ((arg-tns))
       (dolist (arg-type (alien-fun-type-arg-types type))
         (arg-tns (invoke-alien-type-method :arg-tn arg-type arg-state)))
-      (values (my-make-wired-tn 'positive-fixnum 'any-reg nsp-offset)
+      (values (make-wired-tn* 'positive-fixnum any-reg-sc-number nsp-offset)
               (* (max (arg-state-stack-frame-size arg-state) 4) n-word-bytes)
               (arg-tns)
               (invoke-alien-type-method :result-tn
@@ -144,8 +131,8 @@
                                         (make-result-state))))))
 
 (deftransform %alien-funcall ((function type &rest args))
-  (aver (sb!c::constant-lvar-p type))
-  (let* ((type (sb!c::lvar-value type))
+  (aver (sb-c:constant-lvar-p type))
+  (let* ((type (sb-c:lvar-value type))
          (env (make-null-lexenv))
          (arg-types (alien-fun-type-arg-types type))
          (result-type (alien-fun-type-result-type type)))
@@ -154,22 +141,22 @@
     ;; and results.
     (if (or (some #'(lambda (type)
                       (and (alien-integer-type-p type)
-                           (> (sb!alien::alien-integer-type-bits type) 32)))
+                           (> (sb-alien::alien-integer-type-bits type) 32)))
                   arg-types)
             (and (alien-integer-type-p result-type)
-                 (> (sb!alien::alien-integer-type-bits result-type) 32)))
+                 (> (sb-alien::alien-integer-type-bits result-type) 32)))
         (collect ((new-args) (lambda-vars) (new-arg-types))
                  (dolist (type arg-types)
                    (let ((arg (gensym)))
                      (lambda-vars arg)
                      (cond ((and (alien-integer-type-p type)
-                                 (> (sb!alien::alien-integer-type-bits type) 32))
+                                 (> (sb-alien::alien-integer-type-bits type) 32))
                             ;; 64-bit long long types are stored in
                             ;; consecutive locations, endian word order,
                             ;; aligned to 8 bytes.
                             (when (oddp (length (new-args)))
                               (new-args nil))
-                            #!-little-endian
+                            #-little-endian
                             (progn (new-args `(ash ,arg -32))
                                    (new-args `(logand ,arg #xffffffff))
                                    (if (oddp (length (new-arg-types)))
@@ -178,7 +165,7 @@
                                        (new-arg-types (parse-alien-type '(signed 32) env))
                                        (new-arg-types (parse-alien-type '(unsigned 32) env)))
                                    (new-arg-types (parse-alien-type '(unsigned 32) env)))
-                            #!+little-endian
+                            #+little-endian
                             (progn (new-args `(logand ,arg #xffffffff))
                                    (new-args `(ash ,arg -32))
                                    (if (oddp (length (new-arg-types)))
@@ -191,23 +178,23 @@
                             (new-args arg)
                             (new-arg-types type)))))
                  (cond ((and (alien-integer-type-p result-type)
-                             (> (sb!alien::alien-integer-type-bits result-type) 32))
+                             (> (sb-alien::alien-integer-type-bits result-type) 32))
                         (let ((new-result-type
-                               (let ((sb!alien::*values-type-okay* t))
+                               (let ((sb-alien::*values-type-okay* t))
                                  (parse-alien-type
                                   (if (alien-integer-type-signed result-type)
-                                      #!-little-endian
+                                      #-little-endian
                                       '(values (signed 32) (unsigned 32))
-                                      #!+little-endian
+                                      #+little-endian
                                       '(values (unsigned 32) (signed 32))
                                       '(values (unsigned 32) (unsigned 32)))
                                   env))))
                           `(lambda (function type ,@(lambda-vars))
                             (declare (ignore type))
                              (multiple-value-bind
-                               #!-little-endian
+                               #-little-endian
                                (high low)
-                               #!+little-endian
+                               #+little-endian
                                (low high)
                                (%alien-funcall function
                                   ',(make-alien-fun-type
@@ -223,7 +210,7 @@
                               :arg-types (new-arg-types)
                               :result-type result-type)
                            ,@(new-args))))))
-        (sb!c::give-up-ir1-transform))))
+        (sb-c::give-up-ir1-transform))))
 
 (define-vop (foreign-symbol-sap)
   (:translate foreign-symbol-sap)
@@ -236,7 +223,6 @@
   (:generator 2
     (inst li res (make-fixup foreign-symbol :foreign))))
 
-#!+linkage-table
 (define-vop (foreign-symbol-dataref-sap)
   (:translate foreign-symbol-dataref-sap)
   (:policy :fast-safe)
@@ -245,10 +231,9 @@
   (:info foreign-symbol)
   (:results (res :scs (sap-reg)))
   (:result-types system-area-pointer)
-  (:temporary (:scs (non-descriptor-reg)) addr)
   (:generator 2
-    (inst li addr (make-fixup foreign-symbol :foreign-dataref))
-    (loadw res addr)))
+    (inst li res (make-fixup foreign-symbol :foreign-dataref))
+    (loadw res res)))
 
 (define-vop (call-out)
   (:args (function :scs (sap-reg) :target cfunc)
@@ -259,13 +244,17 @@
   (:temporary (:sc any-reg :offset cfunc-offset
                    :from (:argument 0) :to (:result 0)) cfunc)
   (:temporary (:sc control-stack :offset nfp-save-offset) nfp-save)
+  (:temporary (:sc any-reg :offset nl3-offset) tramp) ; = $at
   (:vop-var vop)
   (:generator 0
     (let ((cur-nfp (current-nfp-tn vop)))
       (when cur-nfp
         (store-stack-tn nfp-save cur-nfp))
-      (inst jal (make-fixup "call_into_c" :foreign))
-      (move cfunc function t)
+      ;; (alien-linkage-table-entry-address 0) is "call-into-c" in mips-assem.S
+      (inst lw tramp null-tn (- (alien-linkage-table-entry-address 0) nil-value))
+      (inst nop)
+      (inst jal tramp)
+      (inst move cfunc function)
       (when cur-nfp
         (load-stack-tn cur-nfp nfp-save)))))
 
@@ -310,7 +299,7 @@
   ;; can't we have something like the :ARG-TN methods for all of this
   ;; mess?  -- AB, 2015-Nov-02
   (let* ((parsed-type (parse-alien-type type nil))
-         (alignment-bits (sb!alien::alien-type-alignment parsed-type))
+         (alignment-bits (sb-alien::alien-type-alignment parsed-type))
          (alignment-bytes (truncate alignment-bits n-byte-bits))
          ;; OFFSET is at least 32-bit aligned, we're trying to pick
          ;; out the cases where we need 64-bit alignment.
@@ -319,7 +308,7 @@
      (ecase *backend-byte-order*
        (:big-endian
         (if (alien-integer-type-p parsed-type)
-            (let ((bits (sb!alien::alien-integer-type-bits parsed-type)))
+            (let ((bits (sb-alien::alien-integer-type-bits parsed-type)))
               (let ((byte-offset
                      (cond ((< bits n-word-bits)
                             (- n-word-bytes
@@ -365,15 +354,16 @@ and a pointer to the arguments."
              (let ((offset (* words-processed n-word-bytes)))
                (cond ((not (alien-float-type-p type))
                       (when (and (alien-integer-type-p type)
-                                 (> (sb!alien::alien-integer-type-bits type)
+                                 (> (sb-alien::alien-integer-type-bits type)
                                     n-word-bits)
                                  (oddp words-processed))
                         (pop gprs)
                         (incf words-processed)
                         (incf offset n-word-bytes))
                       (when gprs
-                        (loop repeat words
+                        (loop
                           for gpr = (pop gprs)
+                          repeat words
                           when gpr do
                             (inst sw gpr nsp-tn offset)
                           do
@@ -419,7 +409,7 @@ and a pointer to the arguments."
                         (incf words-processed 2)))
                      (t
                       (bug "Unknown alien floating point type: ~S" type))))))
-        (assemble (segment)
+        (assemble (segment 'nil)
           (mapc #'save-arg argument-types n-argument-words)
           ;; funcall3 (enter-alien-callback, index, args, return-area)
           ;;
@@ -461,7 +451,7 @@ and a pointer to the arguments."
                   (inst lwc1 (make-fpr 1) sp (+ n-callee-register-args-bytes
                                                 n-word-bytes)))))
               ((and (alien-integer-type-p result-type)
-                    (> (sb!alien::alien-integer-type-bits result-type)
+                    (> (sb-alien::alien-integer-type-bits result-type)
                        n-word-bits))
                (inst lw v0 sp n-callee-register-args-bytes)
                (inst lw v1 sp (+ n-callee-register-args-bytes n-word-bytes)))
@@ -481,7 +471,7 @@ and a pointer to the arguments."
       (finalize-segment segment)
       ;; Now that the segment is done, convert it to a static
       ;; vector we can point foreign code to.
-      (let* ((buffer (sb!assem::segment-buffer segment))
+      (let* ((buffer (sb-assem:segment-buffer segment))
              (vector (make-static-vector (length buffer)
                                          :element-type '(unsigned-byte 8)
                                          :initial-contents buffer))
