@@ -17,7 +17,6 @@
 #include "genesis/hash-table.h"
 #include "genesis/static-symbols.h"
 #include "genesis/symbol.h"
-#include "genesis/fdefn.h"
 #include "code.h"
 #include "immobile-space.h"
 #include "queue.h"
@@ -385,7 +384,6 @@ static void sweep_fixedobj_pages()
  * unlike deposit_filler() which tries to be efficient */
 static void clobber_headered_object(lispobj* addr, sword_t nwords)
 {
-    // FIXME: clobbering an object on single-object pages should free entire pages
     page_index_t page = find_page_index(addr);
     if (page < 0) { // code space
 #ifdef LISP_FEATURE_IMMOBILE_SPACE
@@ -403,7 +401,7 @@ static void clobber_headered_object(lispobj* addr, sword_t nwords)
             }
         }
 #endif
-    } else if ((SINGLE_OBJECT_FLAG|page_table[page].type) == (SINGLE_OBJECT_FLAG|PAGE_TYPE_CODE)) {
+    } else if (is_code(page_table[page].type)) {
         // Code pages don't want (0 . 0) fillers, otherwise heap checking
         // gets an error: "object @ 0x..... is non-code on code page"
         addr[0] = make_filler_header(nwords);
@@ -443,6 +441,19 @@ static uword_t sweep(lispobj* where, lispobj* end,
     return 0;
 }
 
+#ifndef LISP_FEATURE_MARK_REGION_GC
+static uword_t sweep_possibly_large(lispobj* where, lispobj* end,
+                                    __attribute__((unused)) uword_t arg)
+{
+    extern void free_large_object(lispobj*, lispobj*);
+    if (page_single_obj_p(find_page_index(where))) {
+        if (!pointer_survived_gc_yet((lispobj)where)) free_large_object(where, end);
+    } else
+        sweep(where, end, arg);
+    return 0;
+}
+#endif
+
 void dispose_markbits() {
     os_deallocate((void*)fullcgcmarks, markbits_size);
     fullcgcmarks = 0; markbits_size = 0;
@@ -458,6 +469,9 @@ void dispose_markbits() {
 
 void execute_full_sweep_phase()
 {
+#ifdef LISP_FEATURE_MARK_REGION_GC
+    lose("Can't do sweep");
+#else
     long words_zeroed[1+PSEUDO_STATIC_GENERATION]; // One count per generation
 
     local_smash_weak_pointers();
@@ -480,7 +494,7 @@ void execute_full_sweep_phase()
             text_page_genmask[find_text_page_index(where)]
                 |= (1 << immobile_obj_gen_bits(where));
 #endif
-    walk_generation(sweep, -1, (uword_t)words_zeroed);
+    walk_generation(sweep_possibly_large, -1, (uword_t)words_zeroed);
     if (gencgc_verbose) {
         fprintf(stderr, "[Sweep phase: ");
         int i;
@@ -489,4 +503,5 @@ void execute_full_sweep_phase()
         fprintf(stderr, " words zeroed]\n");
     }
     dispose_markbits();
+#endif
 }
