@@ -9,7 +9,8 @@
  * files for more information.
  */
 
-#include "sbcl.h"
+#include "genesis/sbcl.h"
+#include "lispobj.h"
 
 #include <stdio.h>
 #include <sys/types.h>
@@ -34,9 +35,12 @@
 #include "lispregs.h"
 #include "interrupt.h"
 #include "thread.h"
+#include "genesis/instance.h"
 #include "genesis/static-symbols.h"
 #include "genesis/primitive-objects.h"
 #include "genesis/gc-tables.h"
+#include "genesis/unwind-block.h"
+#include "genesis/hash-table.h"
 #include "gc.h"
 #include "../../tlsf-bsd/tlsf/tlsf.h"
 extern void* tlsf_control;
@@ -62,8 +66,11 @@ struct crash_preamble {
     uword_t static_nbytes;
     uword_t readonly_start;
     uword_t readonly_nbytes;
+    uword_t permgen_start;
+    uword_t permgen_nbytes;
     uword_t dynspace_start;
-    long dynspace_npages;
+    long dynspace_npages_total;
+    long dynspace_npages_used;
     int card_size;
     int card_table_nbits;
     // fixedobj data dumped: pages, page table
@@ -89,28 +96,57 @@ struct crash_thread_preamble {
 };
 
 // Prevent some mixups in case you add fields to the crash dump
-const int CRASH_PREAMBLE_SIGNATURE =
+const uword_t CRASH_PREAMBLE_SIGNATURE =
     (sizeof (struct crash_preamble) << 16) | sizeof (struct crash_thread_preamble);
 
+__attribute__((unused))
+static void maybe_show_contents(__attribute__((unused)) char *legend,
+                                __attribute__((unused)) void* buf,
+                                __attribute__((unused)) long nbytes)
+{
+#if 0 // def LISP_FEATURE_64_BIT
+    int want = legend && (!strcmp(legend, "preamble") || legend[0] == ' ');
+    if (!want) return;
+    int nwords = nbytes>>WORD_SHIFT;
+    int i;
+    for (i=0;i<nwords;++i)
+        fprintf(stderr, "%s %16lx", (i%8)?"":(i?"\n   ":"   "), ((uword_t*)buf)[i]);
+    putc('\n', stderr);
+#endif
+}
+
 #if defined LISP_FEATURE_X86 || defined LISP_FEATURE_X86_64 // un-tested elsewhere
 #include <errno.h>
-static void checked_write(char *legend, int fd, void* buf, long nbytes)
+struct filewriter { long total; int fd; bool verbose; };
+
+static void checked_write(char *section, struct filewriter* writer, void* buf, long nbytes)
 {
-    ssize_t wrote = write(fd, buf, nbytes);
-    if (wrote != nbytes) lose("short write, errno=%d", errno);
-    fprintf(stderr, "%s: %lx bytes\n", legend, nbytes);
+    char label[8];
+    // I don't know why strncpy() gets a warning here. Isn't truncation exactly the point?
+    // monitor.c:118:5: warning: ‘strncpy’ output truncated before terminating nul copying 8 bytes
+    //   from a string of the same length [-Wstringop-truncation]
+    memset(label, 0, sizeof label);
+    memcpy(label, section, strlen(section)<sizeof label?strlen(section):sizeof label);
+    ssize_t wrote;
+    if (write(writer->fd, label, sizeof label) != sizeof label
+        || (wrote = write(writer->fd, buf, nbytes)) != nbytes)
+        lose("short write, errno=%d", errno);
+    if (writer->verbose) fprintf(stderr, "%s: %lx bytes\n", section, (long)(nbytes + sizeof label));
+    maybe_show_contents(section, buf, nbytes);
+    writer->total += nbytes + sizeof label;
 }
 
 #include "immobile-space.h"
 void save_gc_crashdump(char *pathname,
-                       lispobj* cur_thread_approx_stackptr)
+                       lispobj* cur_thread_approx_stackptr,
+                       bool verbose)
 {
     extern int pin_all_dynamic_space_code;
-    int fd = open(pathname, O_WRONLY|O_CREAT, 0666);
+    int fd = open(pathname, O_WRONLY|O_CREAT|O_TRUNC, 0666);
     struct thread* th;
     int nthreads = 0;
     for_each_thread(th) ++nthreads;
-    fprintf(stderr, "save: %d threads\n", nthreads);
+    if (verbose) fprintf(stderr, "save: %d threads\n", nthreads);
     struct crash_preamble preamble;
     unsigned long nbytes_heap = next_free_page * GENCGC_PAGE_BYTES;
     int nbytes_tls = SymbolValue(FREE_TLS_INDEX,0);
@@ -119,8 +155,11 @@ void save_gc_crashdump(char *pathname,
     preamble.static_nbytes = (uword_t)static_space_free_pointer - STATIC_SPACE_START;
     preamble.readonly_start = READ_ONLY_SPACE_START;
     preamble.readonly_nbytes = (uword_t)read_only_space_free_pointer - READ_ONLY_SPACE_START;
+    preamble.permgen_start = PERMGEN_SPACE_START;
+    preamble.permgen_nbytes = (uword_t)permgen_space_free_pointer - PERMGEN_SPACE_START;
     preamble.dynspace_start = DYNAMIC_SPACE_START;
-    preamble.dynspace_npages = next_free_page;
+    preamble.dynspace_npages_total = dynamic_space_size / GENCGC_PAGE_BYTES;
+    preamble.dynspace_npages_used = next_free_page;
     preamble.card_size = GENCGC_CARD_BYTES;
     preamble.card_table_nbits = gc_card_table_nbits;
     preamble.nthreads = nthreads;
@@ -142,31 +181,38 @@ void save_gc_crashdump(char *pathname,
     preamble.tlsf_control_size = tlsf_size();
     memcpy(preamble.sentinel_block, tlsf_memory_end-3*N_WORD_BYTES, 3*N_WORD_BYTES);
 #endif
+    struct filewriter writer = { .fd = fd, .total = 0, .verbose = verbose };
     // write the preamble and static + readonly spaces
-    checked_write("preamble", fd, &preamble, sizeof preamble);
-    checked_write("static", fd, (char*)STATIC_SPACE_START, preamble.static_nbytes);
-    checked_write("R/O", fd, (char*)READ_ONLY_SPACE_START, preamble.readonly_nbytes);
+    checked_write("preamble", &writer, &preamble, sizeof preamble);
+    checked_write("static", &writer, (char*)STATIC_SPACE_START, preamble.static_nbytes);
+    checked_write("R/O", &writer, (char*)READ_ONLY_SPACE_START, preamble.readonly_nbytes);
+    checked_write("perm", &writer, (char*)PERMGEN_SPACE_START, preamble.permgen_nbytes);
 
     // write the dynamic-space, PTEs, card table
-    checked_write("dynamic", fd, (char*)DYNAMIC_SPACE_START, nbytes_heap);
-    checked_write("PTE", fd, page_table, sizeof (struct page) * next_free_page);
-    checked_write("cardmark", fd, gc_card_mark, 1+gc_card_table_mask);
+    checked_write("dynamic", &writer, (char*)DYNAMIC_SPACE_START, nbytes_heap);
+    checked_write("PTE", &writer, page_table, sizeof (struct page) * next_free_page);
+    checked_write("cardmark", &writer, gc_card_mark, 1+gc_card_table_mask);
+#ifdef LISP_FEATURE_MARK_REGION_GC
+    checked_write("allocated", &writer, allocation_bitmap, bitmap_size(next_free_page));
+    extern line_index_t line_count;
+    checked_write("linemap", &writer, line_bytemap, line_count);
+#endif
 #ifdef LISP_FEATURE_IMMOBILE_SPACE
     int usage = (uword_t)fixedobj_free_pointer - FIXEDOBJ_SPACE_START;
-    checked_write("fixedobj", fd, (char*)FIXEDOBJ_SPACE_START, usage);
+    checked_write("fixedobj", &writer, (char*)FIXEDOBJ_SPACE_START, usage);
     int total_npages = FIXEDOBJ_SPACE_SIZE / IMMOBILE_CARD_BYTES;
-    checked_write("fixedobj_PTE", fd, fixedobj_pages, total_npages * sizeof sizeof(struct fixedobj_page));
+    checked_write("fixedobj_PTE", &writer, fixedobj_pages, total_npages * sizeof sizeof(struct fixedobj_page));
     usage = (uword_t)text_space_highwatermark - TEXT_SPACE_START;
     // write the block_header_t that is just beyond the high water mark
-    checked_write("text", fd, (char*)TEXT_SPACE_START, usage+3*N_WORD_BYTES);
+    checked_write("text", &writer, (char*)TEXT_SPACE_START, usage+3*N_WORD_BYTES);
     total_npages = text_space_size / IMMOBILE_CARD_BYTES;
     int n_bitmap_elts = ALIGN_UP(total_npages, 32) / 32;
-    checked_write("text_gen", fd, text_page_genmask, total_npages); // 1 byte per page
-    checked_write("text_WP", fd, text_page_touched_bits, n_bitmap_elts * sizeof (int));
-    checked_write("TLSF_control", fd, tlsf_control, preamble.tlsf_control_size);
+    checked_write("text_gen", &writer, text_page_genmask, total_npages); // 1 byte per page
+    checked_write("text_WP", &writer, text_page_touched_bits, n_bitmap_elts * sizeof (int));
+    checked_write("TLSF_control", &writer, tlsf_control, preamble.tlsf_control_size);
     int tlsf_memory_size = tlsf_memory_end - (char*)tlsf_mem_start;
     int n_tlsf_pages = tlsf_memory_size / IMMOBILE_CARD_BYTES;
-    checked_write("TLSF_sso", fd, tlsf_page_sso, n_tlsf_pages * sizeof (short));
+    checked_write("TLSF_sso", &writer, tlsf_page_sso, n_tlsf_pages * sizeof (short));
 #endif
     struct crash_thread_preamble thread_preamble;
     for_each_thread(th) {
@@ -176,7 +222,7 @@ void save_gc_crashdump(char *pathname,
         if (ici) {
 #ifdef LISP_FEATURE_C_STACK_IS_CONTROL_STACK
             sp = *os_context_register_addr(threadcontext, reg_SP);
-#else
+#elif defined reg_CSP
             sp = *os_context_register_addr(threadcontext, reg_CSP);
 #endif
         } else if (th->state_word.state == STATE_DEAD) {
@@ -189,7 +235,7 @@ void save_gc_crashdump(char *pathname,
               int n = snprintf(msg, sizeof msg,
                                "thread %p state %d - No stackptr for crash dump\n",
                                th, th->state_word.state);
-              write(2, msg, n);
+              ignore_value(write(2, msg, n));
               _exit(1);
             }
 #ifdef LISP_FEATURE_C_STACK_IS_CONTROL_STACK
@@ -213,27 +259,28 @@ void save_gc_crashdump(char *pathname,
         thread_preamble.control_stack_nbytes = nbytes_control_stack;
         thread_preamble.binding_stack_nbytes = nbytes_binding_stack;
         // write the preamble
-        checked_write("thread", fd, &thread_preamble, sizeof thread_preamble);
+        checked_write("thread", &writer, &thread_preamble, sizeof thread_preamble);
         // write 0 or 1 contexts, control-stack, binding-stack, TLS
-        if (ici) checked_write(" ctxt", fd, threadcontext, preamble.sizeof_context);
+        if (ici) checked_write(" ctxt", &writer, threadcontext, preamble.sizeof_context);
 #ifdef LISP_FEATURE_C_STACK_IS_CONTROL_STACK
-        checked_write(" stack", fd, (char*)sp, nbytes_control_stack);
+        checked_write(" stack", &writer, (char*)sp, nbytes_control_stack);
 #else
-        checked_write(" stack", fd, th->control_stack_start, nbytes_control_stack);
+        checked_write(" stack", &writer, th->control_stack_start, nbytes_control_stack);
 #endif
-        checked_write(" bindings", fd, th->binding_stack_start, nbytes_binding_stack);
-        checked_write(" TLS", fd, th, nbytes_tls);
+        checked_write(" bindings", &writer, th->binding_stack_start, nbytes_binding_stack);
+        checked_write(" TLS", &writer, th, nbytes_tls);
     }
-    checked_write("sig", fd, "SB.Crash", 8); // trailing signature
+    checked_write("sig", &writer, "SB.Crash", 8); // trailing signature
     close(fd);
+    if (verbose) fprintf(stderr, "Total: %ld bytes\n", writer.total);
 }
 #endif
 
 static cmd call_cmd, dump_cmd, print_cmd, quit_cmd, help_cmd;
 static cmd flush_cmd, regs_cmd, exit_cmd;
-static cmd print_context_cmd, pte_cmd, search_cmd;
+static cmd print_context_cmd, pte_cmd, search_cmd, hashtable_cmd;
 static cmd backtrace_cmd, catchers_cmd;
-static cmd threads_cmd, findpath_cmd;
+static cmd threads_cmd, findpath_cmd, layouts_cmd;
 
 extern void gc_stop_the_world(), gc_start_the_world();
 static void suspend_other_threads() {
@@ -252,17 +299,39 @@ static int save_cmd(char **ptr) {
     char *name  = parse_token(ptr);
     if (!name) {
         fprintf(stderr, "Need filename\n");
-        return 1;
+        return 0;
     }
 #if (defined LISP_FEATURE_X86 || defined LISP_FEATURE_X86_64) && defined LISP_FEATURE_SB_THREAD
     suspend_other_threads();
-    save_gc_crashdump(name, (lispobj*)__builtin_frame_address(0));
+    save_gc_crashdump(name, (lispobj*)__builtin_frame_address(0), 1);
     unsuspend_other_threads();
 #else
     fprintf(stderr, "Unimplemented\n");
 #endif
     return 0;
 }
+static int gc_and_save_cmd(char **ptr) {
+    /* The use-case for this is as follows: suppose you're testing a shiny new GC
+     * on a large Lisp application, but gc_and_save crashes 1 time in 10.
+     * How do you effectively debug that if merely getting to the point where you
+     * would run S-L-A-D takes significant time?  Just inject save_gc_crashdump()
+     * into every save at the top of prepare_to_save() or thereabouts, and run
+     * until you collect at least one crash dump. Then iterate on editing the
+     * runtime including the GC, and restarting 'ldb' on the saved state from a
+     * crashed run to try to isolate what went wrong without having to build up
+     * the Lisp heap again from scratch */
+    char *name  = parse_token(ptr);
+    if (!name) {
+        fprintf(stderr, "Need filename\n");
+        return 0;
+    }
+#if defined(LISP_FEATURE_SB_THREAD) && !defined(LISP_FEATURE_WIN32)
+    current_thread = all_threads;
+#endif
+    extern void gc_and_save(char*,bool,bool,bool,bool,int,int);
+    gc_and_save(name, 0, 0, 0, 0, 0, 0); // never returns
+    return 0;
+}
 void list_lisp_threads(int regions) {
     struct thread* th;
     fprintf(stderr, "(thread*,pthread,sb-vm:thread,name)\n");
@@ -373,21 +442,22 @@ static int gc_cmd(char **ptr) {
     return 0;
 }
 
-static int tlsf_cmd(__attribute__((unused)) char **ptr) {
 #ifdef LISP_FEATURE_IMMOBILE_SPACE
+static int tlsf_cmd(__attribute__((unused)) char **ptr) {
     tlsf_dump_pool(tlsf_control, tlsf_mem_start, "/dev/tty");
 #ifdef TLSF_CONFIG_DEBUG
     tlsf_check(tlsf_control);
     tlsf_check_pool(tlsf_mem_start);
-#endif
 #endif
     return 0;
 }
+#endif
 
 static struct cmd {
     char *cmd, *help;
     int (*fn)(char **ptr);
 } supported_cmds[] = {
+    // Commands with no help string are all at-your-own-risk
     {"help", "Display this help information.", help_cmd},
     {"?", "(an alias for help)", help_cmd},
     {"backtrace", "Backtrace up to N frames.", backtrace_cmd},
@@ -399,17 +469,22 @@ static struct cmd {
     {"exit", "Exit this instance of the monitor.", exit_cmd},
     {"findpath", "Find path to an object.", findpath_cmd},
     {"flush", "Flush all temp variables.", flush_cmd},
+    {"hashtable", "Dump a hashtable in detail.", hashtable_cmd},
+    {"layouts", "Dump LAYOUT instances.", layouts_cmd},
     {"print", "Print object at ADDRESS.", print_cmd},
     {"p", "(an alias for print)", print_cmd},
     {"pte", "Page table entry for address", pte_cmd},
     {"quit", "Quit.", quit_cmd},
     {"regs", "Display current Lisp registers.", regs_cmd},
     {"search", "Search heap for object.", search_cmd},
-    {"save", "Produce crashdump", save_cmd},
+    {"save", 0, save_cmd}, // snapshot heap to file ("best effort" though)
+    {"gc_and_save", 0, gc_and_save_cmd},
     {"threads", "List threads", threads_cmd},
-    {"tlsfdump", "Dump TLSF structures", tlsf_cmd},
+#ifdef LISP_FEATURE_IMMOBILE_SPACE
+    {"tlsfdump", 0, tlsf_cmd}, // (unsafely) dump TLSF structures
+#endif
     {"verify", "Check heap invariants", verify_cmd},
-    {"gc", "Collect garbage", gc_cmd},
+    {"gc", 0, gc_cmd}, // (this is not the right way to invoke GC)
     {NULL, NULL, NULL}
 };
 
@@ -549,6 +624,75 @@ print_cmd(char **ptr)
     return 0;
 }
 
+int verify_lisp_hashtable(__attribute__((unused)) struct hash_table* ht,
+                          __attribute__((unused)) FILE* file)
+{
+    int errors = 0;
+#ifdef LISP_FEATURE_64_BIT
+    char *kinds[4] = {"EQ","EQL","EQUAL","EQUALP"};
+    lispobj* data = VECTOR(ht->pairs)->data;
+    uint32_t* hvdata = ht->hash_vector != NIL ?
+      (uint32_t*)VECTOR(ht->hash_vector)->data : 0;
+    struct vector* iv = VECTOR(ht->index_vector);
+    uint32_t* ivdata = (void*)iv->data;
+    uint32_t* nvdata = (void*)VECTOR(ht->next_vector)->data;
+    unsigned ivmask = vector_len(iv) - 1;
+    int hwm = KV_PAIRS_HIGH_WATER_MARK(data);
+    if (file)
+        fprintf(file,
+                "Table %p Kind=%d=%s Weak=%d Count=%d HWM=%d rehash=%d\n",
+                ht, hashtable_kind(ht), kinds[hashtable_kind(ht)],
+                hashtable_weakp(ht)?1:0,
+                (int)fixnum_value(ht->_count), hwm, (int)data[1]);
+    int j;
+    for (j = 1; j <= hwm; j++) {
+        lispobj key = data[2*j];
+        lispobj val = data[2*j+1];
+        if (header_widetag(key) == UNBOUND_MARKER_WIDETAG ||
+            header_widetag(val) == UNBOUND_MARKER_WIDETAG) {
+            if (file) fprintf(file, "[%4d] %12lx %16lx\n", j, key, val);
+            continue;
+        }
+        uint32_t h;
+        if (hvdata && hvdata[j] != 0xFFFFFFFF) {
+            h = hvdata[j];
+            // print the as-stored hash
+            if (file)
+                fprintf(file, "[%4d] %12lx %16lx  %08x %4x (",
+                        j, key, val, h, h & ivmask);
+        } else {
+            // print the hash and then fuzzed hash;
+            lispobj h0 = funcall1(ht->hash_fun, key);
+            h = prefuzz_ht_hash(h0);
+            if (file)
+                fprintf(file, "[%4d] %12lx %16lx %016lx %08x %4x (", j,
+                        key, val, fixnum_value(h0), h, h & ivmask);
+        }
+        // show the chain
+        unsigned cell = ivdata[h & ivmask];
+        while (cell) {
+            if (file) fprintf(file, "%d", cell);
+            lispobj matchp = funcall2(ht->test_fun, key, data[cell*2]);
+            if (matchp != NIL) { if (file) fprintf(file, "\u2713"); break; }
+            if ((cell = nvdata[cell]) != 0 && file) putc(' ', file);
+        }
+        if (!cell) ++errors;
+        if (file) fprintf(file, cell ? ")\n" : ") *\n");
+    }
+#endif
+    return errors;
+}
+static int hashtable_cmd(char **ptr)
+{
+    lispobj obj;
+    if (parse_lispobj(ptr, &obj)) {
+        int errors = verify_lisp_hashtable((void*)native_pointer(obj),
+                                           stdout);
+        if (errors) fprintf(stderr, "Errors: %d\n", errors);
+    }
+    return 0;
+}
+
 static int
 pte_cmd(char **ptr)
 {
@@ -708,6 +852,10 @@ print_context(os_context_t *context)
         brief_print((lispobj)(*os_context_register_addr(context,i)));
 
     }
+#if defined(LISP_FEATURE_DARWIN) && defined(LISP_FEATURE_PPC)
+    printf("DAR:\t\t 0x%08lx\n", (unsigned long)(*os_context_register_addr(context, 41)));
+    printf("DSISR:\t\t 0x%08lx\n", (unsigned long)(*os_context_register_addr(context, 42)));
+#endif
 #ifndef REG_PC
     printf("PC:\t\t  0x%08lx\n", (unsigned long)os_context_pc(context));
 #endif
@@ -801,6 +949,72 @@ catchers_cmd(char __attribute__((unused)) **ptr)
     return 0;
 }
 
+struct layout_collection {
+    struct cons* list;
+    int passno;
+};
+static int count_layout_occurs(lispobj x, struct cons* list)
+{
+    int ct = 0;
+    for ( ; list ; list = (void*)list->cdr ) if (list->car == x) ++ct;
+    return ct;
+}
+
+static uword_t display_layouts(lispobj* where, lispobj* limit, uword_t arg)
+{
+    extern struct vector * classoid_name(lispobj * classoid);
+    struct layout_collection *lc = (void*)arg;
+    where = next_object(where, 0, limit); /* find first marked object */
+    for ( ; where ; where = next_object(where, object_size(where), limit) ) {
+        if (widetag_of(where) == INSTANCE_WIDETAG &&
+            instance_layout(where) != 0 &&
+            layout_depth2_id(LAYOUT(instance_layout(where))) == LAYOUT_LAYOUT_ID) {
+            struct layout* l = (void*)where;
+            struct classoid* c = (void*)native_pointer(l->classoid);
+            if (lc->passno == 1) {
+                // on the first pass, just collect the classoids;
+                struct cons* cons = malloc(sizeof (struct cons));
+                cons->car = (lispobj)c;
+                cons->cdr = (lispobj)lc->list;
+                lc->list = cons;
+            } else {
+                // print some information preceded by a '*' if more than one
+                // layout points to the classoid of this layout
+                int count = count_layout_occurs((lispobj)c, lc->list);
+                struct vector* v = classoid_name((lispobj*)c);
+                char* name =
+                  header_widetag(v->header)==SIMPLE_BASE_STRING_WIDETAG ? (char*)v->data : "?";
+                fprintf(stderr, "%c %p %16" OBJ_FMTX " %16" OBJ_FMTX " %p %" OBJ_FMTX " %s\n",
+                        count>1 ? '*' : ' ', l, l->clos_hash, l->uw_id_word0,
+                        c, l->invalid, name);
+
+            }
+        }
+    }
+   return 0;
+}
+
+static int layouts_cmd(char __attribute__((unused)) **ptr)
+{
+    fprintf(stderr, "Dup, Layout, Hash, ID_Word, Classoid, Invalid, Name\n");
+    struct layout_collection lc;
+    lc.list = 0;
+    for (lc.passno = 1; lc.passno <= 2; ++lc.passno) {
+        walk_generation(display_layouts, -1, (uword_t)&lc);
+#ifdef LISP_FEATURE_IMMOBILE_SPACE
+        display_layouts((lispobj*)FIXEDOBJ_SPACE_START, fixedobj_free_pointer,
+                        (uword_t)&lc);
+#endif
+    }
+    struct cons* l = lc.list;
+    while (l) {
+        struct cons* next = (struct cons*)l->cdr;
+        free(l);
+        l = next;
+    }
+    return 0;
+}
+
 extern FILE *gc_activitylog_file;
 void
 ldb_monitor(void)
@@ -868,38 +1082,22 @@ ldb_monitor(void)
 void gc_stop_the_world() { } // do nothing
 void gc_start_the_world() { } // do nothing
 #include <errno.h>
+#include <setjmp.h>
 #include "core.h"
 struct lisp_startup_options lisp_startup_options;
-char gc_coalesce_string_literals;
 
-void unwind_binding_stack() { lose("Can't unwind binding stack"); }
-FILE *prepare_to_save(__attribute__((unused)) char *filename,
-                      __attribute__((unused)) bool prepend_runtime,
-                      __attribute__((unused)) void **runtime_bytes,
-                      __attribute__((unused)) size_t *runtime_size) {
-    lose("Can't prepare_to_save");
-}
-bool save_runtime_to_filehandle(__attribute__((unused)) FILE *output,
-                                __attribute__((unused)) void *runtime,
-                                __attribute__((unused)) size_t runtime_size,
-                                __attribute__((unused)) int application_type) {
-    lose("Can't save_runtime_to_filehandle");
-}
-bool save_to_filehandle(__attribute__((unused)) FILE *file,
-                        __attribute__((unused)) char *filename,
-                        __attribute__((unused)) lispobj init_function,
-                        __attribute__((unused)) bool make_executable,
-                        __attribute__((unused)) bool save_runtime_options,
-                        __attribute__((unused)) int core_compression_level) {
-    lose("Can't save_to_filehandle");
-}
-
-static size_t checked_read(char *legend, int fd, void* buf, size_t n)
+static size_t checked_read(char *section, int fd, void* buf, size_t n)
 {
-    size_t result = read(fd, buf, n);
-    if (result != n) { lose("read failed, errno=%d", errno); }
-    if (legend) fprintf(stderr, "%s: %lx bytes\n", legend, n);
-    return result;
+    char label[8];
+    if (read(fd, label, sizeof label) != 8 || strncmp(label, section, 8))
+        lose("section messup: %.8s when expecting %s\n", label, section);
+    if (n) {
+        size_t result = read(fd, buf, n);
+        if (result != n) { lose("read failed, errno=%d", errno); }
+        fprintf(stderr, "%s: %zx bytes\n", section, n + sizeof label);
+        maybe_show_contents(section, buf, n);
+    }
+    return n + sizeof label;
 }
 
 char *pagetypedesc(int type)
@@ -924,7 +1122,6 @@ int load_gc_crashdump(char* pathname)
     int fd;
     os_context_t *contexts[10], *context;
     struct thread* threads = 0;
-    struct thread dummy;
     fd = open(pathname, O_RDONLY);
     if (fd < 0) {
         fprintf(stderr, "can't open %s\n", pathname);
@@ -934,7 +1131,8 @@ int load_gc_crashdump(char* pathname)
     struct crash_thread_preamble thread_preamble;
     checked_read("preamble", fd, &preamble, sizeof preamble);
     printf("static=%"OBJ_FMTX" nbytes=%x\n", preamble.static_start, (int)preamble.static_nbytes);
-    printf("heap_start=%"OBJ_FMTX" npages=%d\n", preamble.dynspace_start, (int)preamble.dynspace_npages);
+    printf("heap_start=%"OBJ_FMTX" npages=%d\n", preamble.dynspace_start,
+           (int)preamble.dynspace_npages_used);
     // pin_dynspace_code is for display only. It gets recomputed as the
     // logical OR of all threads' values of *GC-PIN-CODE-PAGES*.
     printf("sprof_enabled=%d pin_dynspace_code=%d packages=%p\n",
@@ -948,21 +1146,43 @@ int load_gc_crashdump(char* pathname)
     if (preamble.card_size != GENCGC_CARD_BYTES)
         lose("Can't load crashdump: memory parameters differ");
     gc_card_table_nbits = preamble.card_table_nbits;
-    gc_allocate_ptes();
-    next_free_page = preamble.dynspace_npages;
     // static + readonly
     checked_read("static", fd, (char*)STATIC_SPACE_START, preamble.static_nbytes);
     static_space_free_pointer = (lispobj*)(STATIC_SPACE_START + preamble.static_nbytes);
-    if (preamble.readonly_nbytes) {
-      // READ_ONLY_SPACE_START = preamble.readonly_start;
-      os_alloc_gc_space(READ_ONLY_CORE_SPACE_ID, 0, (char*)preamble.readonly_start,
-                        ALIGN_UP(preamble.readonly_nbytes, 4096));
-      checked_read("R/O", fd, (char*)preamble.readonly_start, preamble.readonly_nbytes);
+    if (!preamble.readonly_nbytes) {
+        checked_read("R/O", fd, 0, 0);
+    } else {
+        void* actual =
+            os_alloc_gc_space(READ_ONLY_CORE_SPACE_ID, 0, (char*)preamble.readonly_start,
+                              ALIGN_UP(preamble.readonly_nbytes, 4096));
+        if (actual != (void*)preamble.readonly_start)
+            fprintf(stderr, "WARNING: wanted R/O space @ %p but got %p\n",
+                    (char*)preamble.readonly_start, actual);
+        checked_read("R/O", fd, (char*)preamble.readonly_start, preamble.readonly_nbytes);
+#ifndef READ_ONLY_SPACE_START /* if non-constant */
+        READ_ONLY_SPACE_START = preamble.readonly_start;
+        READ_ONLY_SPACE_END = READ_ONLY_SPACE_START + preamble.readonly_nbytes;
+        read_only_space_free_pointer = (lispobj*)READ_ONLY_SPACE_END;
+#endif
+    }
+    if (!preamble.permgen_nbytes) {
+        checked_read("perm", fd, 0, 0);
+    } else {
+        void* actual =
+            os_alloc_gc_space(PERMGEN_CORE_SPACE_ID, 0, (char*)preamble.permgen_start,
+                              ALIGN_UP(preamble.permgen_nbytes, 4096));
+        if (actual != (void*)preamble.permgen_start)
+            lose("Couldn't map permgen as required (%p)", (char*)preamble.permgen_start);
+        checked_read("perm", fd, (char*)preamble.permgen_start, preamble.permgen_nbytes);
+        PERMGEN_SPACE_START = preamble.permgen_start;
+        permgen_space_free_pointer = (lispobj*)(preamble.permgen_start + preamble.permgen_nbytes);
     }
-    read_only_space_free_pointer = (lispobj*)(READ_ONLY_SPACE_START + preamble.readonly_nbytes);
     //
+    gc_allocate_ptes();
+    dynamic_space_size = preamble.dynspace_npages_total * GENCGC_PAGE_BYTES;
+    next_free_page = preamble.dynspace_npages_used;
     DYNAMIC_SPACE_START = preamble.dynspace_start;
-    long dynspace_nbytes = preamble.dynspace_npages * GENCGC_PAGE_BYTES;
+    long dynspace_nbytes = preamble.dynspace_npages_used * GENCGC_PAGE_BYTES;
     char* dynspace = os_alloc_gc_space(DYNAMIC_CORE_SPACE_ID, 0, (char*)preamble.dynspace_start,
                                        DEFAULT_DYNAMIC_SPACE_SIZE);
     if (dynspace != (char*)preamble.dynspace_start)
@@ -972,8 +1192,16 @@ int load_gc_crashdump(char* pathname)
     fprintf(stderr, "snapshot: %"PRIdPTR" pages in use (%ld bytes)\n",
             next_free_page, dynspace_nbytes);
     checked_read("PTE", fd, page_table, sizeof (struct page) * next_free_page);
-    recompute_gen_bytes_allocated();
     checked_read("cardmark", fd, gc_card_mark, 1+gc_card_table_mask);
+    // mrgc_init needs to know the dynamic space size so that's why this is delayed
+    // until after reading the crash preamble.
+    gc_init();
+#ifdef LISP_FEATURE_MARK_REGION_GC
+    checked_read("allocated", fd, allocation_bitmap, bitmap_size(next_free_page));
+    extern line_index_t line_count;
+    checked_read("linemap", fd, line_bytemap, line_count);
+#endif
+    recompute_gen_bytes_allocated();
     print_generation_stats();
 #ifdef LISP_FEATURE_IMMOBILE_SPACE
     extern void gc_init_immobile(),
@@ -1040,22 +1268,37 @@ int load_gc_crashdump(char* pathname)
         *os_context_register_addr(context,reg_SP) = (os_context_register_t)stackptr;
 #endif
         gc_assert(*os_context_register_addr(context,reg_SP) == (os_context_register_t)stackptr);
-#else
+#elif defined reg_CSP
         *os_context_register_addr(context, reg_CSP) = (uword_t)stackptr;
 #endif
         checked_read(" stack", fd, stackptr, thread_preamble.control_stack_nbytes);
         checked_read(" bindings", fd, th->binding_stack_start, thread_preamble.binding_stack_nbytes);
-        // Skip over the initial words of the thread structure that was saved
-        // in the file, so that binding_stack_start remains as is in the
-        // newly allocated structure. The last word is the only one we want to keep.
-        int skip = sizeof dummy-N_WORD_BYTES;
-        checked_read(0, fd, &dummy, skip);
-        checked_read(" TLS", fd, &th->lisp_thread, preamble.tls_size-skip);
+        {
+        // Capture the 'struct thread' portion of TLS so that things like
+        // binding stack pointer, and prev,next links stay as they are,
+        // after copying them back. This way the visual display of number of bytes
+        // read comes out correctly
+        struct thread preserve;
+        memcpy(&preserve, th, sizeof *th);
+        checked_read(" TLS", fd, th, preamble.tls_size);
+        // Copy _into_ preserve the values of the allocation regions so that when
+        // copied back into 'th' we have access to the regions as they were at dump.
+        preserve.boxed_tlab = th->boxed_tlab;
+        preserve.cons_tlab = th->cons_tlab;
+        preserve.mixed_tlab = th->mixed_tlab;
+        preserve.symbol_tlab = th->symbol_tlab; // not used yet
+        preserve.sys_mixed_tlab = th->sys_mixed_tlab;
+        preserve.sys_cons_tlab = th->sys_cons_tlab;
+        memcpy(th, &preserve, sizeof *th - N_WORD_BYTES);
+        }
         write_TLS(FREE_INTERRUPT_CONTEXT_INDEX, make_fixnum(1), th);
-        struct thread_instance* instance = (void*)(th->lisp_thread - INSTANCE_POINTER_LOWTAG);
-        lispobj name = instance->_name;
-        char* cname = gc_managed_addr_p(name) && simple_base_string_p(name)
-                      ? vector_sap(name) : 0;
+        char* cname = "(no lisp thread)";
+        if (th->lisp_thread) {
+            struct thread_instance* instance = (void*)(th->lisp_thread - INSTANCE_POINTER_LOWTAG);
+            lispobj name = instance->_name;
+            cname = gc_managed_addr_p(name) && simple_base_string_p(name)
+                    ? vector_sap(name) : 0;
+        }
         fprintf(stderr, "thread @ %p originally %p, %d bind_stk words, %d val_stk words '%s'\n",
                 th, (void*)thread_preamble.address,
                 (int)(thread_preamble.binding_stack_nbytes>>WORD_SHIFT),
@@ -1101,6 +1344,8 @@ int load_gc_crashdump(char* pathname)
     return 0;
 }
 
+char *sbcl_runtime;
+jmp_buf ldb_toplevel;
 int main(int argc, char *argv[], char **envp)
 {
     extern void calc_asm_routine_bounds();
@@ -1110,10 +1355,11 @@ int main(int argc, char *argv[], char **envp)
     }
     bool have_hardwired_spaces = os_preinit(argv, envp);
     allocate_lisp_dynamic_space(have_hardwired_spaces);
-    gc_init();
     load_gc_crashdump(argv[1]);
     calc_asm_routine_bounds();
     gencgc_verbose = 1;
+    if (setjmp(ldb_toplevel) != 0)
+      fprintf(stderr, "Back in ldb, hope everything's Ok\n");
     ldb_monitor();
 }
 #endif
